// ======================================================================
//
// MeshAppear.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "MeshAppear.h"

#include "Camera.h"
#include "Extents.h"
#include "Gl.h"
#include "Iff.h"
#include "Material.h"
#include "Random.h"

// ======================================================================

const Tag TAG_FVRT = TAG(F,V,R,T);
const Tag TAG_VTNM = TAG(V,T,N,M);
const Tag TAG_MATS = TAG(M,A,T,S);
const Tag TAG_0013 = TAG(0,0,1,3);

// ======================================================================

UVertex         MeshAppear::vertexBuffer[MAX_VERTICES_PER_POLY];

int             MeshAppear::polygonSelectStart;
int             MeshAppear::polygonSelectStop;

// ======================================================================
// debugging routine that solid fills materials with random colors

static int pickMaterialNum(int num)
{
	if (DFlag(DF_E_RANDOM_SOLID))
		num = (0x7f << 24) | ((rand() & 0xff) << 16) | ((rand() & 0xff) << 8) | ((rand() & 0xff) << 0);

	return num;
}

static const Material *pickMaterial(const Material *mat)
{
	if (DFlag(DF_E_RANDOM_SOLID))
	{
		MaterialList::release(mat);
		mat = MaterialList::fetch((0x7f << 24) | ((rand() & 0xff) << 16) | ((rand() & 0xff) << 8) | ((rand() & 0xff) << 0));
	}

	return mat;
}

// ======================================================================

MeshAppear::MeshAppear(const char *iffname, Iff &iff, real newScale)
: Appear(AppearIdMesh, iffname, newScale), center()
{
	material = NULL;
	materialCount = 0;
	faceVertex = NULL;

	switch (iff.currentName())
	{
		case TAG_0013:
			iff.seekForm(TAG_0013);
			load_0013(iffname, iff);
			iff.exitForm();
			break;

		case TAG_0012:
			iff.seekForm(TAG_0012);
			load_0012(iffname, iff);
			iff.exitForm();
			break;

		case TAG_0011:
			iff.seekForm(TAG_0011);
			load_0011(iffname, iff);
			iff.exitForm();
			break;

		case TAG_0010:
			iff.seekForm(TAG_0010);
			load_0010(iffname, iff);
			iff.exitForm();
			break;

		case TAG_0009:
			iff.seekForm(TAG_0009);
			load_0009(iffname, iff);
			iff.exitForm();
			break;

		default:
			DEBUG_FATAL(true, ("MeshAppear::MeshAppear unknown version in %s", iffname));
			break;
	}

	// sanity checking on the loaded mesh
	DEBUG_FATAL(radius < 0, ("MeshAppear::MeshAppear radius < 0"));

	setExtents(NEW Extents(center, radius));

#ifdef _DEBUG
	for (int i = 0; i < normalCount; ++i)
	{
		real r;

		r = abs(CONST_REAL(1.0) - normal[i].magnitude());
		WARNING(r > CONST_REAL(0.01), ("Mesh:: bad normal %s %d/%d %8.6f\n", iffname, i, normalCount, normal[i].magnitude()));
	}
#endif
}

// ----------------------------------------------------------------------

MeshAppear::~MeshAppear(void)
{
	int i;

	for (i = 0; i < faceCount; ++i)
	{
		MaterialList::release(face[i].material);
		face[i].material = NULL;
		MaterialList::release(face[i].solidMaterial);
		face[i].solidMaterial = NULL;
	}
	delete [] face;

	for (i = 0; i < materialCount; ++i)
	{
		DEBUG_FATAL(!material, ("MeshAppear::~ no material list"));
		MaterialList::release(material[i]);
		material[i] = NULL;
	}
	delete [] material;

	delete [] faceVertex;
	delete [] vertex;
	delete [] normal;
	delete [] friendlyName;
}

// ----------------------------------------------------------------------

void MeshAppear::load_0013(const char *meshName, Iff &iff)
{
	UNREF (meshName);

	int        i;

	iff.seekChunk(TAG_NAME);
	friendlyName = MEM_OWN(iff.read_char(Iff::ReadRest));
	iff.exitChunk();

	// load materials
	if (iff.seekForm(TAG_MATS, Iff::FlagOptional))
	{
		materialCount = iff.blockCount();
		material = NEW const Material *[materialCount];

		for (i = 0; i < materialCount; ++i)
			material[i] = MaterialList::fetch(&iff);

		iff.exitForm();
	}

	// vertex
	iff.seekChunk(TAG_VERT);
	vertexCount = iff.chunkLength(sizeof(Vector::Save));
	DEBUG_FATAL(vertexCount == 0, ("MeshAppear::load_0013 vertices %d", vertexCount));
	vertex = MEM_OWN(iff.read_floatVector(vertexCount));
	iff.exitChunk();

	// scale the vertices
	if (scale != CONST_REAL(1))
		for (i = 0; i < vertexCount; ++i)
			vertex[i] *= scale;

	// read the list of vertex normals
	iff.seekChunk(TAG_VTNM);
	normalCount = iff.chunkLength(sizeof(Vector::Save));
	normal = MEM_OWN(iff.read_floatVector(normalCount));
	iff.exitChunk();

	// face vertex structure
	iff.seekChunk(TAG_FVRT);
	faceVertexCount = iff.chunkLength(sizeof(int32) + sizeof(int32) + sizeof(float) + sizeof(float));
	faceVertex    = NEW FaceVertex[faceVertexCount];
	for (i = 0; i < faceVertexCount; ++i)
	{
		faceVertex[i].vertexIndex = iff.read_int32();
		faceVertex[i].normalIndex = iff.read_int32();
		if (faceVertex[i].normalIndex < 0)
			Fatal ("MeshAppear::load_0013: faceVertex[%d].normalIndex < 0", i);
		faceVertex[i].u           = iff.read_float();
		faceVertex[i].v           = iff.read_float();
	}
	iff.exitChunk();

	// read the faces
	iff.seekChunk(TAG_FACE);
	faceCount = iff.chunkLength(sizeof(int32) + sizeof(float) + sizeof(int32) + sizeof(int32) + sizeof(int32) + sizeof(int32) + sizeof(int32));
	face    = NEW Face[faceCount];
	for (i = 0; i < faceCount; ++i)
	{
		face[i].normalIndex      = iff.read_int32();
		if (face[i].normalIndex < 0)
			Fatal ("MeshAppear::load_0013: face[%d].normalIndex < 0", i);
		face[i].planeD           = iff.read_float() * scale;
		face[i].material         = pickMaterial(MaterialList::fetch(material[iff.read_int32()]));
		face[i].faceVertex       = faceVertex+iff.read_int32();
		face[i].vertexCount      = static_cast<u_int8>(iff.read_int32());
		face[i].flags            = static_cast<u_int8>(iff.read_int32());
		face[i].solidMaterial    = MaterialList::fetch(material[iff.read_int32()]);
	}
	iff.exitChunk();

	if (iff.seekChunk(TAG_BSP, Iff::FlagOptional))
	{
		WARNING(true, ("MeshAppear::load %s has a BSP\n", meshName));
		iff.exitChunk();	
	}
	
	// centroid
	iff.seekChunk(TAG_CNTR);
	iff.read_floatVector(1, &center);
	center *= scale;
	iff.exitChunk();

	// radius of sphere around centroid
	iff.seekChunk(TAG_RADI);
	radius = iff.read_float() * scale;
	iff.exitChunk();
}

// ----------------------------------------------------------------------

void MeshAppear::load_0012(const char *meshName, Iff &iff)
{
	UNREF (meshName);

	int        i;

	iff.seekChunk(TAG_NAME);
	friendlyName = MEM_OWN(iff.read_char(Iff::ReadRest));
	iff.exitChunk();

	// load anonymous materials
	if (iff.seekForm(TAG_MATS, Iff::FlagOptional))
	{
		materialCount = iff.blockCount();
		delete [] material;
		material = NEW const Material *[materialCount];

		for (i = 0; i < materialCount; ++i)
			material[i] = MaterialList::fetch(&iff);

		iff.exitForm();
	}

	// vertex
	iff.seekChunk(TAG_VERT);
	vertexCount = iff.chunkLength(sizeof(Vector::Save));
	DEBUG_FATAL(vertexCount == 0, ("MeshAppear::load_0012 vertices %d", vertexCount));
	vertex = MEM_OWN(iff.read_floatVector(vertexCount));
	iff.exitChunk();

	// scale the vertices
	if (scale != CONST_REAL(1))
		for (i = 0; i < vertexCount; ++i)
			vertex[i] *= scale;

	// read the list of vertex normals
	iff.seekChunk(TAG_VTNM);
	normalCount = iff.chunkLength(sizeof(Vector::Save));
	normal = MEM_OWN(iff.read_floatVector(normalCount));
	iff.exitChunk();

	// face vertex structure
	iff.seekChunk(TAG_FVRT);
	faceVertexCount = iff.chunkLength(sizeof(int32) + sizeof(int32) + sizeof(float) + sizeof(float));
	faceVertex    = NEW FaceVertex[faceVertexCount];
	for (i = 0; i < faceVertexCount; ++i)
	{
		faceVertex[i].vertexIndex = iff.read_int32();
		faceVertex[i].normalIndex = iff.read_int32();
		if (faceVertex[i].normalIndex < 0)
			Fatal ("MeshAppear::load_0012: faceVertex[%d].normalIndex < 0", i);
		faceVertex[i].u           = iff.read_float();
		faceVertex[i].v           = iff.read_float();
	}
	iff.exitChunk();

	// read the faces
	iff.seekChunk(TAG_FACE);
	faceCount = iff.chunkLength(sizeof(int32) + sizeof(float) + sizeof(int32) + sizeof(int32) + sizeof(int32) + sizeof(int32) + sizeof(int32));
	face    = NEW Face[faceCount];
	for (i = 0; i < faceCount; ++i)
	{
		face[i].normalIndex      = iff.read_int32();
		if (face[i].normalIndex < 0)
			Fatal ("MeshAppear::load_0012: face[%d].normalIndex < 0", i);
		face[i].planeD           = iff.read_float() * scale;
		face[i].material         = MaterialList::fetch(pickMaterialNum(iff.read_int32()));
		face[i].faceVertex       = faceVertex+iff.read_int32();
		face[i].vertexCount      = static_cast<u_int8>(iff.read_int32());
		face[i].flags            = static_cast<u_int8>(iff.read_int32());
		face[i].solidMaterial    = MaterialList::fetch(iff.read_int32());
	}
	iff.exitChunk();

	if (iff.seekChunk(TAG_BSP, Iff::FlagOptional))
	{
		WARNING(true, ("MeshAppear::load %s has a BSP\n", meshName));
		iff.exitChunk();	
	}
	
	// centroid
	iff.seekChunk(TAG_CNTR);
	iff.read_floatVector(1, &center);
	center *= scale;
	iff.exitChunk();

	// radius of sphere around centroid
	iff.seekChunk(TAG_RADI);
	radius = iff.read_float() * scale;
	iff.exitChunk();
}

// ----------------------------------------------------------------------

void MeshAppear::load_0011(const char *meshName, Iff &iff)
{
	UNREF (meshName);

	int        i;
	int        vertexNormalCount;
	Vector    *vertexNormal;
	int        faceNormalCount;
	Vector    *faceNormal;

	WARNING(true, ("MeshAppear::load %s version 0011\n", meshName));

	iff.seekChunk(TAG_NAME);
	friendlyName = MEM_OWN(iff.read_char(Iff::ReadRest));
	iff.exitChunk();

	// vertex
	iff.seekChunk(TAG_VERT);
	vertexCount = iff.chunkLength(sizeof(Vector::Save));
	DEBUG_FATAL(vertexCount == 0, ("MeshAppear::load_0011 vertices %d", vertexCount));
	vertex = MEM_OWN(iff.read_floatVector(vertexCount));
	iff.exitChunk();

	// scale the vertices
	if (scale != CONST_REAL(1))
		for (i = 0; i < vertexCount; ++i)
			vertex[i] *= scale;

 	// normals
	iff.seekChunk(TAG_NORM);
	faceNormalCount = iff.chunkLength(sizeof(Vector::Save));
	faceNormal = MEM_OWN(iff.read_floatVector(faceNormalCount));
	iff.exitChunk();

	// read the list of vertex normals
	iff.seekChunk(TAG_VTNM);
	vertexNormalCount = iff.chunkLength(sizeof(Vector::Save));
	vertexNormal = MEM_OWN(iff.read_floatVector(vertexNormalCount));
	iff.exitChunk();

	// build the unified normal list
	normalCount = faceNormalCount + vertexNormalCount;
	normal = NEW Vector[normalCount];
	memcpy(normal, faceNormal, faceNormalCount * sizeof(*normal));
	memcpy(normal+faceNormalCount, vertexNormal, vertexNormalCount * sizeof(*normal));
	delete [] faceNormal;
	faceNormal = NULL;
	delete [] vertexNormal;
	vertexNormal = NULL;

	// face vertex structure
	iff.seekChunk(TAG_FVRT);
	faceVertexCount = iff.chunkLength(sizeof(int32) + sizeof(int32) + sizeof(float) + sizeof(float));
	faceVertex    = NEW FaceVertex[faceVertexCount];
	for (i = 0; i < faceVertexCount; ++i)
	{
		faceVertex[i].vertexIndex = iff.read_int32();
		faceVertex[i].normalIndex = faceNormalCount + iff.read_int32();
		faceVertex[i].u           = iff.read_float();
		faceVertex[i].v           = iff.read_float();
	}
	iff.exitChunk();

	// read the faces
	iff.seekChunk(TAG_FACE);
	faceCount = iff.chunkLength(sizeof(int32) + sizeof(float) + sizeof(int32) + sizeof(int32) + sizeof(int32) + sizeof(int32) + sizeof(int32));
	face    = NEW Face[faceCount];
	for (i = 0; i < faceCount; ++i)
	{
		face[i].normalIndex      = iff.read_int32();
		face[i].planeD           = iff.read_float() * scale;
		face[i].material         = MaterialList::fetch(pickMaterialNum(iff.read_int32()));
		face[i].faceVertex       = faceVertex+iff.read_int32();
		face[i].vertexCount      = static_cast<u_int8>(iff.read_int32());
		face[i].flags            = static_cast<u_int8>(iff.read_int32());
		face[i].solidMaterial    = MaterialList::fetch(iff.read_int32());
	}
	iff.exitChunk();

	if (iff.seekChunk(TAG_BSP, Iff::FlagOptional))
	{
		WARNING(true, ("MeshAppear::load %s has a BSP\n", meshName));
		iff.exitChunk();	
	}
	
	// centroid
	iff.seekChunk(TAG_CNTR);
	iff.read_floatVector(1, &center);
	center *= scale;
	iff.exitChunk();

	// radius of sphere around centroid
	iff.seekChunk(TAG_RADI);
	radius = iff.read_float() * scale;
	iff.exitChunk();
}

// ----------------------------------------------------------------------

void MeshAppear::load_0010(const char *meshName, Iff &iff)
{
	UNREF (meshName);

	int        i;
	int        vertexNormalCount;
	Vector    *vertexNormal;
	int        faceNormalCount;
	Vector    *faceNormal;

	WARNING(true, ("MeshAppear::load %s version 0010\n", meshName));

	iff.seekChunk(TAG_NAME);
	friendlyName = MEM_OWN(iff.read_char(Iff::ReadRest));
	iff.exitChunk();

	// vertex
	iff.seekChunk(TAG_VERT);
	vertexCount = iff.chunkLength(sizeof(Vector::Save));
	DEBUG_FATAL(vertexCount == 0, ("MeshAppear::load_0010 vertices %d", vertexCount));
	vertex = MEM_OWN(iff.read_floatVector(vertexCount));
	iff.exitChunk();

	// scale the vertices
	for (i = 0; i < vertexCount; ++i)
		vertex[i] *= scale;

 	// normals
	iff.seekChunk(TAG_NORM);
	faceNormalCount = iff.chunkLength(sizeof(Vector::Save));
	faceNormal = MEM_OWN(iff.read_floatVector(faceNormalCount));
	iff.exitChunk();

	// read the list of vertex normals
	iff.seekChunk(TAG_VTNM);
	vertexNormalCount = iff.chunkLength(sizeof(Vector::Save));
	vertexNormal = MEM_OWN(iff.read_floatVector(vertexNormalCount));
	iff.exitChunk();

	// build the unified normal list
	normalCount = faceNormalCount + vertexNormalCount;
	normal = NEW Vector[normalCount];
	memcpy(normal, faceNormal, faceNormalCount * sizeof(*normal));
	memcpy(normal+faceNormalCount, vertexNormal, vertexNormalCount * sizeof(*normal));
	delete [] faceNormal;
	faceNormal = NULL;
	delete [] vertexNormal;
	vertexNormal = NULL;

	// face vertex structure
	iff.seekChunk(TAG_FVRT);
	faceVertexCount = iff.chunkLength(sizeof(int32) + sizeof(int32) + sizeof(float) + sizeof(float));
	faceVertex      = NEW FaceVertex[faceVertexCount];
	for (i = 0; i < faceVertexCount; ++i)
	{
		faceVertex[i].vertexIndex = iff.read_int32();
		faceVertex[i].normalIndex = faceNormalCount + iff.read_int32();
		faceVertex[i].u           = iff.read_float();
		faceVertex[i].v           = iff.read_float();
	}
	iff.exitChunk();

	// read the faces
	iff.seekChunk(TAG_FACE);
	faceCount = iff.chunkLength(sizeof(int32) +sizeof(float) +sizeof(int32) +sizeof(int32) +sizeof(int32) +sizeof(int32));
	face    = NEW Face[faceCount];
	for (i = 0; i < faceCount; ++i)
	{
		face[i].normalIndex      = iff.read_int32();
		face[i].planeD           = iff.read_float() * scale;
		face[i].material         = MaterialList::fetch(pickMaterialNum(iff.read_int32()));
		face[i].faceVertex       = faceVertex+iff.read_int32();
		face[i].vertexCount      = static_cast<u_int8>(iff.read_int32());
		face[i].flags            = static_cast<u_int8>(iff.read_int32());
		face[i].solidMaterial    = MaterialList::fetch(0x7fffffff);
	}
	iff.exitChunk();

	// centroid
	iff.seekChunk(TAG_CNTR);
	iff.read_floatVector(1, &center);
	center *= scale;
	iff.exitChunk();

	// radius of sphere around centroid
	iff.seekChunk(TAG_RADI);
	radius = iff.read_float() * scale;
	iff.exitChunk();
}

// ----------------------------------------------------------------------

void MeshAppear::load_0009(const char *meshName, Iff &iff)
{
	UNREF (meshName);

	int  i, j, faceNormalCount;

	WARNING(true, ("MeshAppear::load %s version 0009\n", meshName));

	iff.seekChunk(TAG_NAME);
	friendlyName = MEM_OWN(iff.read_char(Iff::ReadRest));
	iff.exitChunk();

	// vertex
	iff.seekChunk(TAG_VERT);
	vertexCount = iff.chunkLength(sizeof(Vector::Save));
	DEBUG_FATAL(vertexCount == 0, ("MeshAppear::load_0009 vertices %d", vertexCount));
	vertex = MEM_OWN(iff.read_floatVector(vertexCount));
	iff.exitChunk();

 	// normals
	iff.seekChunk(TAG_NORM);
	faceNormalCount = iff.chunkLength(sizeof(Vector::Save));
	normalCount = faceNormalCount + vertexCount;
	normal = NEW Vector[normalCount];
	iff.read_floatVector(faceNormalCount, normal);
	for (i = faceNormalCount; i < normalCount; ++i)
		normal[i].makeZero();
	iff.exitChunk();

	// face vertex structure
	iff.seekChunk(TAG_FVRT);
	faceVertexCount = iff.chunkLength(sizeof(int32) + sizeof(int32) + sizeof(float) + sizeof(float));
	faceVertex    = NEW FaceVertex[faceVertexCount];
	for (i = 0; i < faceVertexCount; ++i)
	{
		faceVertex[i].vertexIndex = iff.read_int32();
		iff.read_int32();	// dummy holder for the normal index
		faceVertex[i].normalIndex = faceNormalCount + faceVertex[i].vertexIndex;
		faceVertex[i].u           = iff.read_float();
		faceVertex[i].v           = iff.read_float();
	}
	iff.exitChunk();

	// read the faces
	iff.seekChunk(TAG_FACE);
	faceCount = iff.chunkLength(sizeof(int32) +sizeof(float) +sizeof(int32) +sizeof(int32) +sizeof(int32) +sizeof(int32));
	face    = NEW Face[faceCount];
	for (i = 0; i < faceCount; ++i)
	{
		face[i].normalIndex     = iff.read_int32();
		face[i].planeD          = iff.read_float();
		face[i].material        = MaterialList::fetch(pickMaterialNum(iff.read_int32()));
		face[i].faceVertex      = faceVertex+iff.read_int32();
		face[i].vertexCount     = static_cast<u_int8>(iff.read_int32());
		face[i].flags           = static_cast<u_int8>(iff.read_int32());
		face[i].solidMaterial   = MaterialList::fetch(0x7fffffff);
	}
	iff.exitChunk();

	// centroid
	iff.seekChunk(TAG_CNTR);
	iff.read_floatVector(1, &center);
	iff.exitChunk();

	// radius of sphere around centroid
	iff.seekChunk(TAG_RADI);
	radius = iff.read_float();
	iff.exitChunk();

	// calculate the direction of all the vertex normals by adding the face normal to each vertex's normal in the face
	for (i = 0; i < faceCount; ++i)
		for (j = 0; j < face[i].vertexCount; ++j)
			normal[face[i].faceVertex[j].normalIndex] += normal[face[i].normalIndex];

	// now normalize all the vertex normals, while avoiding divide-by-zero
	for (i = faceNormalCount; i < normalCount; ++i)
		if (!normal[i].normalize())
			normal[i].makeZero();
}

// ----------------------------------------------------------------------

void MeshAppear::write(const char *filename) const
{
	int   i, j, matCount;
	Iff   iff(1024);
	int32 i32;
	float f;
	
	const Material **mat;

	// build unique material list if one did not already exist
	if (!material)
	{
		mat = NEW const Material *[faceCount * 2];
		matCount = 0;

		for (i = 0; i < faceCount; ++i)
		{
			// add normal materials
			for (j = 0; j < matCount && face[i].material->getCrcName()->compareCrcString(*(mat[j]->getCrcName())) != 0; ++j)
				;
			if (j == matCount)
				mat[matCount++] = MaterialList::fetch(face[i].material);

			// add solid materials
			for (j = 0; j < matCount && face[i].solidMaterial->getCrcName()->compareCrcString(*(mat[j]->getCrcName())) != 0; ++j)
				;
			if (j == matCount)
				mat[matCount++] = MaterialList::fetch(face[i].solidMaterial);
		}
	}
	else
	{
		mat      = material;
		matCount = materialCount;
	}

	iff.appendForm(TAG_MESH);
	iff.appendForm(TAG_0013);

	iff.appendChunk(TAG_NAME, friendlyName, strlen(friendlyName)+1);

	iff.appendForm(TAG_MATS);
	for (i = 0; i < matCount; ++i)
		if (mat[i]->getName())
			iff.appendChunk(TAG_NAME, mat[i]->getName(), strlen(mat[i]->getName())+1);
		else
		{
			DEBUG_FATAL(true, ("MeshAppear::write material without a name"));
		}
	iff.exitForm();

	// vertex
	iff.appendChunk(TAG_VERT);
	for (i = 0; i < vertexCount; ++i)
	{
		f = static_cast<float>(vertex[i].x);
		iff.appendToChunk(&f, sizeof(f));		
		f = static_cast<float>(vertex[i].y);
		iff.appendToChunk(&f, sizeof(f));		
		f = static_cast<float>(vertex[i].z);
		iff.appendToChunk(&f, sizeof(f));		
	}

	// read the list of normals
	iff.appendChunk(TAG_VTNM);
	for (i = 0; i < normalCount; ++i)
	{
		f = static_cast<float>(normal[i].x);
		iff.appendToChunk(&f, sizeof(f));		
		f = static_cast<float>(normal[i].y);
		iff.appendToChunk(&f, sizeof(f));		
		f = static_cast<float>(normal[i].z);
		iff.appendToChunk(&f, sizeof(f));		
	}

	// face vertex structure
	iff.appendChunk(TAG_FVRT);
	for (i = 0; i < faceVertexCount; ++i)
	{
		i32 = faceVertex[i].vertexIndex;
		iff.appendToChunk(&i32, sizeof(i32));

		i32 = faceVertex[i].normalIndex;
		iff.appendToChunk(&i32, sizeof(i32));

		f   = faceVertex[i].u;
		iff.appendToChunk(&f, sizeof(f));

		f   = faceVertex[i].v;
		iff.appendToChunk(&f, sizeof(f));
	}

	// read the faces
	iff.appendChunk(TAG_FACE);
	for (i = 0; i < faceCount; ++i)
	{
		i32 = face[i].normalIndex;
		iff.appendToChunk(&i32, sizeof(i32));

		f   = face[i].planeD;
		iff.appendToChunk(&f, sizeof(f));

		for (i32 = 0; i32 < matCount && face[i].material->getCrcName()->compareCrcString(*(mat[i32]->getCrcName())) != 0; ++i32)
			;	
		DEBUG_FATAL(i32 == matCount, ("MeshAppear::write mat not found"));
		iff.appendToChunk(&i32, sizeof(i32));
						
		i32 = face[i].faceVertex - faceVertex;
		iff.appendToChunk(&i32, sizeof(i32));

		i32 = face[i].vertexCount;
		iff.appendToChunk(&i32, sizeof(i32));

		i32 = face[i].flags;
		iff.appendToChunk(&i32, sizeof(i32));

		for (i32 = 0; i32 < matCount && face[i].solidMaterial->getCrcName()->compareCrcString(*(mat[i32]->getCrcName())) != 0; ++i32)
			;	
		DEBUG_FATAL(i32 == matCount, ("MeshAppear::write solid mat not found"));
		iff.appendToChunk(&i32, sizeof(i32));
	}

	// centroid
	iff.appendChunk(TAG_CNTR);
	f = static_cast<float>(center.x);
	iff.appendToChunk(&f, sizeof(f));		
	f = static_cast<float>(center.y);
	iff.appendToChunk(&f, sizeof(f));		
	f = static_cast<float>(center.z);
	iff.appendToChunk(&f, sizeof(f));		

	// radius of sphere around centroid
	iff.appendChunk(TAG_RADI);
	f = static_cast<float>(radius);
	iff.appendToChunk(&f, sizeof(f));		

	iff.exitForm();
	iff.exitForm();

	iff.write(filename);

	if (mat != material)
	{
		for (i = 0; i < matCount; ++i)
			MaterialList::release(mat[i]);
		delete [] mat;
	}
}

// ----------------------------------------------------------------------

const Appear *MeshAppear::getRenderAppear(Camera &camera, const Object *object) const
{
	Vector v;

	object->rotateTranslateO2W(center, &v);

	if (camera.testSphereVisibility(v, radius))
		return Appear::getRenderAppear(camera, object);

	++Camera::renderStats->objectsCulled;
	return NULL;
}

// ----------------------------------------------------------------------

void MeshAppear::renderFaces(Face *face, int faceCount) const
{
	int              i;
	FaceVertex      *faceVertex;
	UVertex         *pvi;

	// for all faces
	for (; faceCount--; ++face)
	{	
		++Camera::renderStats->polygonTotalCount;

		// build the output polygon vertices from the polygon struct
		for (i = face->vertexCount, faceVertex = face->faceVertex, pvi = vertexBuffer; i--; ++faceVertex, ++pvi)
		{
			pvi->vertexIndex = faceVertex->vertexIndex;
			pvi->normalIndex = DFlag(DF_E_FLAT_SHADE) ? face->normalIndex : faceVertex->normalIndex;
			pvi->u           = faceVertex->u;
			pvi->v           = faceVertex->v;						
			pvi->r           = CONST_REAL(1);
			pvi->g           = CONST_REAL(1);
			pvi->b           = CONST_REAL(1);
			pvi->a           = CONST_REAL(1);
		}

		if (face->flags & 2)
			Gl::setLightingEnabled(false);

		// get the material
		if (DFlag(DF_E_DEBUG_MAT))
			Gl::setMaterial(MaterialList::debugMaterial()->getStaticMaterial());
		else
			if (DFlag(DF_E_SOLID_FILL))
				Gl::setMaterial(face->solidMaterial->getStaticMaterial());
			else
				Gl::setMaterial(face->material->getStaticMaterial());

		if (face->flags & 2)
			Gl::setLightingEnabled(true);

		Gl::drawTriangleList(vertexBuffer, face->vertexCount);
	}
}

// ----------------------------------------------------------------------

void MeshAppear::render(Camera &camera, const Xform &objectToWorld) const
{
	// set up the object to world transform
	Gl::setObjectToWorld(objectToWorld);

	// push these vertices into the GL
	if (Gl::setVertexList(vertex, vertexCount, normal, normalCount))
	{
		// render the polygons
		if (DKeyIsContext(DK_E_POLYGON_SELECT))
		{
			int start, stop;
	
			start = clamp(0, polygonSelectStart, faceCount);
			stop  = clamp(0, polygonSelectStop,  faceCount);

			if (stop > start)
				renderFaces(face+start, stop-start);
		}
		else
			renderFaces(face, faceCount);

		if (DFlag(DF_E_WIRE_FRAME))
			wireFrame();

		if (DFlag(DF_E_WIRE_AXIS))
			wireAxis();

		if (DFlag(DF_E_WIRE_POLYGON_NORMALS))
			wirePolygonNormals();

		if (DFlag(DF_E_WIRE_VERTEX_NORMALS))
			wireVertexNormals();
	}

	Appear::render(camera, objectToWorld);
}

// ----------------------------------------------------------------------

void MeshAppear::wireFrame(void) const
{
	Gl::setMaterial(MaterialList::white());

	int start, stop, i, j, k;
	UVertex uv[2];

	if (DKeyIsContext(DK_E_POLYGON_SELECT))
	{
		start = clamp(0, polygonSelectStart, faceCount);
		stop  = clamp(0, polygonSelectStop,  faceCount);
	}
	else
	{
		start = 0;
		stop = faceCount;
	}

	for (i = 0; i < 2; ++i)
	{
		uv[i].normalIndex = -1;

		uv[i].u = CONST_REAL(0);
		uv[i].v = CONST_REAL(0);

		uv[i].r = CONST_REAL(1);
		uv[i].g = CONST_REAL(1);
		uv[i].b = CONST_REAL(1);
		uv[i].a = CONST_REAL(1);
	}

	for (i = start; i < stop; ++i)
		for (k = face[i].vertexCount-1, j = 0; j < face[i].vertexCount; k = j++)
		{
			uv[0].vertexIndex = face[i].faceVertex[k].vertexIndex;
			uv[1].vertexIndex = face[i].faceVertex[j].vertexIndex;

			Gl::drawLineList(uv, 2);
		}
}

// ----------------------------------------------------------------------

void MeshAppear::wireAxis(void) const
{
	Vector v[4] =
	{
		Vector::zero,
		Vector::unitX,
		Vector::unitY,
		Vector::unitZ
	};

	// scale the axies to the radius of the mesh
	v[1] *= radius;
	v[2] *= radius;
	v[3] *= radius;

	Gl::setMaterial(MaterialList::red());
	Gl::line(v[0], v[1]);

	Gl::setMaterial(MaterialList::green());
	Gl::line(v[0], v[2]);

	Gl::setMaterial(MaterialList::blue());
	Gl::line(v[0], v[3]);
}

// ----------------------------------------------------------------------

void MeshAppear::wirePolygonNormals(void) const
{
	int    start, stop, i, j;
	Vector v0, v1;
	
	if (DKeyIsContext(DK_E_POLYGON_SELECT))
	{
		start = clamp(0, polygonSelectStart, faceCount);
		stop  = clamp(0, polygonSelectStop,  faceCount);
	}
	else
	{
		start = 0;
		stop  = faceCount;
	}

	for (i = start; i < stop; ++i)
	{
		// build the normal start point
		for (v0 = Vector::zero, j = 0; j < face[i].vertexCount; ++j)
			v0 += vertex[face[i].faceVertex[j].vertexIndex];
		v0 /= face[i].vertexCount;

		// build the normal end point
		v1.add(v0, normal[face[i].normalIndex]);
		
		// draw the line for the normal
		Gl::setMaterial(MaterialList::green());
		Gl::line(v0, v1);

		// plot a point at the tip to indiate direction
		Gl::setMaterial(MaterialList::red());
		Gl::point(v1);
	}
}

// ----------------------------------------------------------------------

void MeshAppear::wireVertexNormals(void) const
{
	int    start, stop, i;
	Vector v0, v1;

	if (DKeyIsContext(DK_E_POLYGON_SELECT))
	{
		start = clamp(0, polygonSelectStart, faceCount);
		stop  = clamp(0, polygonSelectStop,  faceCount);

		if (start < faceCount && stop > start)
		{
			start = face[start].faceVertex - faceVertex;
			stop  = (face[stop-1].faceVertex+face[stop-1].vertexCount) - faceVertex;
		}
	}
	else
	{
		start = 0;
		stop  = faceVertexCount;
	}

	for (i = start; i < stop; ++i)
	{
		// build the normal start point
		v0 = vertex[faceVertex[i].vertexIndex];
		
		// get the normal
		v1 = normal[faceVertex[i].normalIndex];
		
		// make sure the normal was valid
		if (abs(v1.magnitude() - CONST_REAL(1)) > CONST_REAL(0.01))
		{
			// draw an obviously bad normal
			v1 = v0;
			if (!v1.normalize())
				v1 = Vector::unitZ;
			v1 *= CONST_REAL(5);
			v1 += v0;

			Gl::setMaterial(MaterialList::red());			
			Gl::line(v0, v1);
		}
		else
		{
			// build the normal end point
			v1 += v0;

			// draw the normal			
			Gl::setMaterial(MaterialList::blue());			
			Gl::line(v0, v1);

			// plot a point at the end to indicate direction
			Gl::setMaterial(MaterialList::red());			
			Gl::point(v1);
		}
	}
}

// ======================================================================
